/*
 * cntlr.c
 * wifi controller core
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <easy/easy.h>

#include <timer_impl.h>
#include <cmdu.h>
#include <1905_tlvs.h>
#include <i1905_wsc.h>
#include <map2.h>
#include <map_module.h>
#include <uci.h>

#include "utils/debug.h"
#include "utils/utils.h"
#include "cntlr.h"
#include "allsta.h"
#include "cntlr_ubus.h"
#include "cntlr_map.h"
#include "cntlr_cmdu.h"
#include "steer_module.h"
#include "cntlr_acs.h"

#define map_plugin	"ieee1905.map"


extern bool waitext;

/* find interface by macaddress - no radio argument */
struct netif_iface *find_interface_by_mac_nor(struct controller *c,
		uint8_t *hwaddr)
{
	struct netif_iface *p = NULL;	/* fh anf bk iface */
	struct netif_radio *r = NULL;
	struct node *n = NULL;

	list_for_each_entry(n, &c->nodelist, list) {
		list_for_each_entry(r, &n->radiolist, list) {
			list_for_each_entry(p, &r->iflist, list) {
				if (!memcmp(p->bssid, hwaddr, 6))
					return p;
			}
		}
	}

	return NULL;
}

/* find interface by macaddress */
struct netif_iface *find_interface_by_mac(struct controller *c,
		struct netif_radio *r, uint8_t *hwaddr)
{
	struct netif_iface *p = NULL;

	list_for_each_entry(p, &r->iflist, list) {
		if (!memcmp(p->bssid, hwaddr, 6))
			return p;
	}

	return NULL;
}

/* find radio by ssid */
struct netif_radio *find_radio_by_ssid(struct controller *c,
		struct node *n, char *ssid)
{
	struct netif_radio *r = NULL;

	list_for_each_entry(r, &n->radiolist, list) {

		struct netif_iface *p = NULL;

		list_for_each_entry(p, &r->iflist, list) {
			if (!memcmp(p->ssid, ssid, 33))
				return r;
		}
	}

	return NULL;
}

/* find netif by ssid */
struct netif_iface *find_interface_by_ssid(struct controller *c,
		struct node *n, char *ssid)
{
	struct netif_radio *r = NULL;

	list_for_each_entry(r, &n->radiolist, list) {

		struct netif_iface *p = NULL;

		list_for_each_entry(p, &r->iflist, list) {
			if (!memcmp(p->ssid, ssid, 33))
				return p;
		}
	}

	return NULL;
}

/* find radio by node */
struct netif_radio *find_radio_by_node(struct controller *c, struct node *n,
		uint8_t *radio_mac)
{
	struct netif_radio *p = NULL;

	list_for_each_entry(p, &n->radiolist, list) {
		if (!memcmp(p->macaddr, radio_mac, 6))
			return p;
	}

	return NULL;
}

/* finds radio struct from interface macaddr */
struct netif_radio *find_radio_by_bssid(struct controller *c, uint8_t *bssid)
{
	struct node *n = NULL;
	struct netif_radio *r = NULL;
	struct netif_iface *p = NULL;

	list_for_each_entry(n, &c->nodelist, list) {
		list_for_each_entry(r, &n->radiolist, list) {
			list_for_each_entry(p, &r->iflist, list) {
				if (!memcmp(p->bssid, bssid, 6))
					return r;
			}
		}
	}

	return NULL;
}

/* find link by macaddress */
struct netif_link *find_link_by_mac(struct controller *c, uint8_t *upstream, uint8_t *downstream)
{
	struct netif_link *l = NULL;

	list_for_each_entry(l, &c->linklist, list) {
		if (!memcmp(l->upstream->bssid, upstream, 6) && !memcmp(l->downstream->bssid, downstream, 6))
			return l;
	}

	return NULL;
}

/* find node by macaddress */
struct node *find_node_by_mac(struct controller *c, uint8_t *mac)
{
	struct node *p = NULL;

	list_for_each_entry(p, &c->nodelist, list) {
		if (!memcmp(p->alid, mac, 6))
			return p;
	}

	return NULL;
}

/* find sta by macaddress */
struct sta *cntlr_find_sta(struct controller *c, uint8_t *mac)
{
	struct sta *s = NULL;

	list_for_each_entry(s, &c->stalist, list) {
		if (!memcmp(s->macaddr, mac, 6))
			return s;
	}

	return NULL;
}

struct bcnreq *cntlr_find_bcnreq(struct controller *c, uint8_t *sta, uint8_t *alid)
{
	struct bcnreq *br = NULL;

	dbg("%s: --->\n", __func__);

	list_for_each_entry(br, &c->bcnreqlist, list) {
		if (!memcmp(br->sta_mac, sta, 6) && !memcmp(br->agent_mac, alid, 6))
			return br;
	}

	return NULL;
}

#if 0
/* find node by macaddress */
struct netif_iface *cntlr_get_fbss_by_mac(struct controller *c, struct node *n,
		uint8_t *mac)
{
	struct netif_iface *p;

	list_for_each_entry(p, &n->iflist, list) {
		if (!memcmp(p->bssid, mac, 6))
			return p;
	}

	return NULL;
}
#endif
/* find fbss based on macaddr */
struct netif_iface *cntlr_iterate_fbss(struct controller *c, uint8_t *mac)
{
	struct node *n = NULL;
	struct netif_radio *r = NULL;
	struct netif_iface *p = NULL;

	list_for_each_entry(n, &c->nodelist, list) {
		list_for_each_entry(r, &n->radiolist, list) {
			list_for_each_entry(p, &r->iflist, list) {
				if (!memcmp(p->bssid, mac, 6))
					return p;
			}
		}
	}

	return NULL;
}

/* find node based on bssid */
struct node *find_node_by_iface(struct controller *c, uint8_t *bssid)
{
	struct node *n = NULL;

	list_for_each_entry(n, &c->nodelist, list) {
		struct netif_radio *r;

		list_for_each_entry(r, &n->radiolist, list) {
			struct netif_iface *p;

			list_for_each_entry(p, &r->iflist, list) {
				if (!memcmp(p->bssid, bssid, 6))
					return n;
			}
		}
	}

	return NULL;
}

#if 0
/* find node by ip address */
static struct node *find_node_by_ip(struct controller *c, const char *ip)
{
	struct node *p;
	struct in_addr ipn;

	if (ip && strlen(ip) && !inet_aton(ip, &ipn)) {
		warn("Invalid ipaddr: %s\n", ip);
		return NULL;
	}

	list_for_each_entry(p, &c->nodelist, list) {
		if (!memcmp(&p->ipaddr, &ipn, sizeof(struct in_addr)))
			return p;
	}

	return NULL;
}
#endif

struct node_policy *agent_find_policy(struct controller *c, uint8_t *agent)
{
	struct node_policy *a = NULL;

	list_for_each_entry(a, &c->cfg.nodelist, list) {
		if (!memcmp(agent, a->agent_id, 6))
			return a;
	}

	return NULL;
}

struct radio_policy *agent_find_radio_policy(struct controller *c, uint8_t *radio_mac)
{
	struct node_policy *a = NULL;

	list_for_each_entry(a, &c->cfg.nodelist, list) {
		struct radio_policy *r = NULL;

		list_for_each_entry(r, &a->radiolist, list) {
			if (!memcmp(radio_mac, r->macaddr, 6))
				return r;
		}
	}


	return NULL;
}

#if 0
/* find node by any of its fh-bssid */
struct node *get_node_by_bssid(struct controller *c, unsigned char *bssid)
{
	struct node *n;
	struct netif_iface *p;

	list_for_each_entry(n, &c->nodelist, list) {
		list_for_each_entry(p, &n->iflist, list) {
			if (memcmp(bssid, p->bssid, 6))
				continue;

			return n;
		}
	}

	return NULL;
}
#endif

static void cntlr_log_nodes(struct controller *c)
{
	struct node *n;
	int i = 0;

	list_for_each_entry(n, &c->nodelist, list) {
		cntlr_dbg("  %d | agent = %p,  hwaddr = '"MACFMT"')\n",
				i++, n, MAC2STR(n->alid));
	}
}


#if 0
static int forall_node_update_neighbors(struct controller *c)
{
	return 0;
}
#endif

void cntlr_update_sta_steer_counters(struct controller *c,
				     uint8_t *sta_mac,
				     uint8_t *src_bssid,
				     uint8_t *dst_bssid,
				     enum steer_method method,
				     enum steer_trigger trigger)
{
	trace("%s:--->\n", __func__);

	struct steer_attempt *a;
	struct sta *s = cntlr_find_sta(c, sta_mac);

	if (!s) {
		dbg("|%s:%d| Unrecognized STA "MACFMT", skip!\n",
		    __func__, __LINE__, MAC2STR(sta_mac));
		return;
	}

	if (s->num_steer_attempts >= MAX_LOG_STEER_ATTEMPTS) {
		int i;

		for (i = 0; i < MAX_LOG_STEER_ATTEMPTS - 1; i++) {
			memcpy(&s->attempts[i], &s->attempts[(i+1)],
					sizeof(struct steer_attempt));
		}
		a = &s->attempts[MAX_LOG_STEER_ATTEMPTS - 1];
	} else {
		a = &s->attempts[s->num_steer_attempts];
	}

	/* Update SteeringHistory */
	timestamp_update(&a->time);

	if (src_bssid)
		memcpy(a->src_bssid, src_bssid, 6);

	if (dst_bssid)
		memcpy(a->dst_bssid, dst_bssid, 6);

	a->trigger = trigger;
	a->method = method;
	a->duration = 0;
	a->answered = false;
	a->auto_bssid = dst_bssid ? false : true;

	/* Update SteeringSummaryStats */
	/*TODO: split counter for opportunity & mandate */
	s->stats.btm_attempt_cnt++;
	/* Record tsp for most recent steer attempt */
	timestamp_update(&s->stats.last_steer_time);
	s->num_steer_attempts += 1;
}

static int cntlr_steer_sta(struct controller *c, struct sta *s,
			   struct steer_target *to, uint32_t mode,
			   uint32_t reason)
{
	struct cmdu_buff *cmdu;

	trace("%s:--->\n", __func__);

	if (!to || hwaddr_is_zero(to->bssid)) {
		dbg("%s: steer verdict = ok, but target AP = NULL!\n", __func__);
		return 0;
	}

	if (!memcmp(to->bssid, s->bssid, 6)) {
		s->stats.no_candidate_cnt++;
		dbg("%s: " MACFMT " connected to best AP! No steer needed.\n",
		    __func__);
		return 0;
	}

	dbg("%s: Try to steer " MACFMT " from " MACFMT " to " MACFMT "\n",
	     __func__, MAC2STR(s->macaddr), MAC2STR(s->bssid), MAC2STR(to->bssid));

	UNUSED(mode);	//TODO
	UNUSED(reason);

	cmdu = cntlr_gen_client_steer_request(c, s->fh->agent->alid,
					s->bssid, 0,
					1, (uint8_t (*)[6])s->macaddr,
					1, (uint8_t (*)[6])to->bssid,
					1); /* mandate */
	if (!cmdu) {
		warn("%s: Failed to generate cmdu for steering sta!\n", __func__);
		return -1;
	}

	cntlr_update_sta_steer_counters(c, s->macaddr, s->bssid, to->bssid,
					METHOD_BTM_REQ, TRIGGER_LINK_QUALITY);
	send_cmdu(c, cmdu);
	cmdu_free(cmdu);

	return 0;
}

#define STEER_RCPI_DIFFSNR 8
static void cntlr_configure_steer(struct controller *c, struct sta *s)
{
	struct netif_radio *r;
	struct radio_policy *rp = NULL;
	struct steer_config scfg = {};

	r = find_radio_by_bssid(c, s->bssid);
	if (!r)
		return;

	rp = agent_find_radio_policy(c, r->macaddr);
	if (!rp)
		return;

	/* RCPI threshold */
	if (rp->rcpi_threshold > 0)
		scfg.rcpi_threshold = rp->rcpi_threshold; /* band dependent */
	else
		scfg.rcpi_threshold = DEFAULT_RCPI_THRESHOLD;

	scfg.rcpi_hysteresis = 5; /* TODO: unused */
	scfg.rcpi_diffsnr = STEER_RCPI_DIFFSNR; /* TODO: config */

	cntlr_configure_steer_module(c, &scfg);
}

static void cntlr_try_steer_sta(struct controller *c, struct sta *s)
{
	trace("%s:--->\n", __func__);
	struct steer_sta candidate = {
		.s = s,
		.nbrlist = NULL,
		.bcnlist = &s->bcnlist,
		.best = NULL,
	};
	int ret;

	/* check if sta should be steered */
	ret = cntlr_maybe_steer_sta(c, &candidate);
	if (ret) {
		dbg("cntlr_maybe_steer_sta() ret = %d\n", ret);
		return;
	}

	switch (candidate.verdict) {
	case STEER_VERDICT_OK:
		if (!timestamp_expired(&s->stats.last_steer_time, STEER_ATTEMPT_MIN_INTV)) {
			dbg("%s: last steer attempt < %us ago; skip steering\n",
			    __func__, STEER_ATTEMPT_MIN_INTV / 1000);
			return;
		}

		cntlr_steer_sta(c, s, candidate.best, candidate.mode, candidate.reason);
		break;
	case STEER_VERDICT_NOK:
		return;
	case STEER_VERDICT_MAYBE:
		/* TODO: check next steer-control ? */
		break;
	case STEER_VERDICT_EXCLUDE:
		/* TODO: exclude this STA from subsequent steer attempts */
		break;
	default:
		break;
	}

}

static void cntlr_bcn_metrics_parse(struct uloop_timeout *t)
{
	trace("%s:--->\n", __func__);

	struct sta *s = container_of(t, struct sta, bcn_metrics_timer);
	struct node *n = s->fh->agent;
	struct controller *c = n->cntlr;
	struct netif_iface *bss = NULL;
	struct bcn_metrics *b = NULL, *tmp;

	dbg("%s: STA " MACFMT" connected to " MACFMT " in Node " MACFMT"\n",
	    __func__, MAC2STR(s->macaddr), MAC2STR(s->bssid), MAC2STR(n->alid));

	list_for_each_entry_safe(b, tmp, &s->bcnlist, list) {
		dbg("bcn-report from " MACFMT "\n", MAC2STR(b->bssid));

		/* Skip entry not in our network */
		bss = cntlr_iterate_fbss(c, b->bssid);
		if (!bss) {
			list_del(&b->list);
			free(b);
			s->num_bcn_metrics--;
			dbg("Delete alien entry "MACFMT"\n", MAC2STR(b->bssid));
		}
	}

	if (c->cfg.enable_sta_steer) {
		cntlr_configure_steer(c, s);
		/* configure individually for each STA */
		cntlr_try_steer_sta(c, s);
	}

	dbg("%s exiting\n", __func__);
}

static void cntlr_init_sta_steer_counters(struct sta *s)
{
	memset(&s->stats, 0, sizeof(struct steer_stats));

	/* TODO: implement stats marked as NO_DATA */
	s->stats.blacklist_attempt_cnt = STEER_STATS_NO_DATA;
	s->stats.blacklist_success_cnt = STEER_STATS_NO_DATA;
	s->stats.blacklist_failure_cnt = STEER_STATS_NO_DATA;
	//s->stats.no_candidate_cnt = 0;
	//s->stats.btm_attempt_cnt = 0;
	//s->stats.btm_success_cnt = 0;
	//s->stats.btm_failure_cnt = 0;
	//s->stats.btm_query_resp_cnt = 0;
	//s->stats.last_steer_time = 0;
}

struct sta *cntlr_add_sta(struct controller *c, uint8_t *macaddr)
{
	struct sta *s;

	s = cntlr_find_sta(c, macaddr);
	if (s)
		return s;

	s = calloc(1, sizeof(struct sta));
	if (!s)
		return NULL;

	INIT_LIST_HEAD(&s->bcnlist);
	s->num_bcn_metrics = 0;
	INIT_LIST_HEAD(&s->unassoclist);
	memcpy(s->macaddr, macaddr, 6);
	list_add(&s->list, &c->stalist);
	s->bcn_metrics_timer.cb = cntlr_bcn_metrics_parse;
	cntlr_init_sta_steer_counters(s);

	return s;
}

#if 0
static void forall_but_node_restrict_sta(struct controller *c, struct node *ex,
							uint8_t *sta)
{
}

static int cntlr_monitor_sta(struct controller *c, uint8_t *sta,
						struct node *sta_node)
{
	char sta_macstr[12 + 1] = {0};
	struct node *n;

	btostr((unsigned char *)sta, 6, sta_macstr);
	list_for_each_entry(n, &c->nodelist, list) {
		/* skip monitoring on sta's current node */
		if (n == sta_node)
			continue;

		if (list_empty(&n->iflist))
			continue;

	}

	return 0;
}
#endif

#if 0
static void forall_node_get_usta_metrics(struct controller *c)
{
	struct sta *s;

	list_for_each_entry(s, &c->stalist, list) {
		;
		struct cmdu_buff *cmdu;

		/* TODO: */
		//cmdu = cntlr_gen_unassoc_sta_metric_query(c, s->fh->agent->alid, s->macaddr,
		// opclass, num_metrics, metrics);
		if (!cmdu)
			continue;

		send_cmdu(c, cmdu);
		cmdu_free(cmdu);
	}
}
#endif

static void forall_node_get_sta_metrics(struct controller *c)
{
	struct sta *s = NULL;

	list_for_each_entry(s, &c->stalist, list) {
		struct cmdu_buff *cmdu;

		cmdu = cntlr_gen_sta_metric_query(c, s->fh->agent->alid, s->macaddr);
		if (!cmdu)
			continue;

		send_cmdu(c, cmdu);
		cmdu_free(cmdu);
	}
}

#if 0
static void forall_node_get_fhinfo(struct controller *c)
{
	//struct node *n;

}

static void node_get_monitor_sta(void *cntlr, void *resp, int len, void *cookie)
{
}

static int cntlr_handle_sta_rssi_low(struct controller *c, struct node *sta_n,
					uint8_t *sta, int lowrssi)
{
	struct node *n;
	char sta_macstr[12 + 1] = {0};
	struct active_sta *s_entry;

	if (hwaddr_is_zero(sta))
		return 0;

	s_entry = as_lookup(c->as_table, sta);
	if (!s_entry)
		return -1;

	/* reset lowrssi and best_nbr */
	s_entry->lowrssi = lowrssi;
	s_entry->best_nbr_rssi = -127;
	memset(s_entry->best_nbr, 0, 6);

	/* get this sta's monitored data from other nodes,
	 * if any available.
	 */
	btostr((unsigned char *)sta, 6, sta_macstr);
	list_for_each_entry(n, &c->nodelist, list) {
		if (n == sta_n)
			continue;   /* skip sta's currently assoc'd node */

	}

	cntlr_dbg("%s: After doing get monitor sta .......\n", __func__);
	if (hwaddr_is_zero(s_entry->best_nbr))
		return 0;

#define lowrssi_thresh	-65
#define diffrssi	9

	if (lowrssi < lowrssi_thresh &&
		(s_entry->best_nbr_rssi != -127 &&
		 s_entry->best_nbr_rssi - lowrssi >= diffrssi)) {

		/* found a better neighbor for this lowrssi sta */
		//TODO

		/* 1. assoc_control STA on all nodes but the best_nbr node */
		/* 2. send steer mandate */
	}

	return 0;
}

static bool endpoint_hwaddr(struct controller *c, unsigned char *sta)
{
	struct node *n;

	list_for_each_entry(n, &c->nodelist, list) {
		if (!memcmp(n->alid, sta, 6))
			return false;
	}

	return true;
}

static void cntlr_handle_assoclist_resp(void *cntlr,
					void *resp,
					int len,
					void *cookie)
{
	struct controller *c = (struct controller *)cntlr;
	struct netif_fhbss *p, *fh = (struct netif_fhbss *)cookie;
	unsigned char rbytes[sizeof(struct cmd_get_assoclist)] = {0};
	unsigned char *bssid = fh->bssid;
	struct cmd_get_assoclist *info;
	bool found = false;
	int i, num_sta;
	struct node *n;

	/* cntlr_dbg("%s(): received: %s\n", __func__, (char *)resp); */
	strtob(resp, len, rbytes);
	info = (struct cmd_get_assoclist *)rbytes;
	num_sta = ntohl(info->num);

	/* TODO: include node pointer in netif_fhbss, so that
	 * we don't have to parse for node from fh.
	 */
	list_for_each_entry(n, &c->nodelist, list) {
		list_for_each_entry(p, &n->iflist, list) {
			if (p == fh) {
				found = true;
				break;
			}
		}
	}

	if (!found)
		n = NULL;

	for (i = 0; i < num_sta; i++) {
		unsigned char *macaddr = &info->sta[6 * i];

		if (!endpoint_hwaddr(c, macaddr))
			continue;

		as_insert(c->as_table, macaddr, bssid);

		if (n) {
			cntlr_dbg("Start Monitoring for STA " MACFMT " ....\n",
						MAC2STR(macaddr));
			cntlr_monitor_sta(c, macaddr, n);
		}
	}
}
#endif

#if 0
/* XXX: now doing for only the first fh-iface of a node */
static void forall_node_get_assoclist(struct controller *c)
{
	struct node *n;

	list_for_each_entry(n, &c->nodelist, list) {
		struct netif_fhbss *p;
		struct cmd_get_assoclist req;
		char reqbuf[2 * sizeof(struct cmd_get_assoclist) + 1] = {0};

		if (list_empty(&n->iflist))
			continue;

		p = list_first_entry(&n->iflist,
				struct netif_fhbss, list);
		if (!p)
			continue;

		memset(&req, 0, sizeof(req));
		strncpy(req.ifname, p->ifname, 16);
		btostr((unsigned char *)&req, sizeof(req), reqbuf);


		if (ret)
			cntlr_dbg("Failed to get assoclist (ret = %d)\n", ret);
	}
}
#endif

int invoke_disconnect_sta(struct node *n, struct netif_iface *p,
		unsigned char *hwaddr)
{
	return 0;
}

int invoke_disconnect_sta_by_bssid(struct controller *c, unsigned char *bssid,
						unsigned char *hwaddr)
{
	return 0;
#if 0
	int ret = 0;
	struct node *n;
	struct netif_fhbss *p;

	n = get_node_by_bssid(c, bssid);
	if (!n)
		return -1;

	list_for_each_entry(p, &n->iflist, list) {
		if (memcmp(bssid, p->bssid, 6))
			continue;

		invoke_disconnect_sta(n, p, hwaddr);

		break;
	}

	return ret;
#endif
}


#if 0
static void disconnect_repeated_macs(struct controller *c, struct node *n,
							unsigned char *sta)
{
	struct active_sta *as;
	int hidx = as_hash(sta);

	hlist_for_each_entry(as, &c->as_table[hidx], hlist) {
		if (as->hwaddr[3] ^ sta[3] ||
				as->hwaddr[4] ^ sta[4] ||
				as->hwaddr[5] ^ sta[5])
			continue;

		if (as->hwaddr[1] != n->mobility_domain[0] ||
				as->hwaddr[2] != n->mobility_domain[1])
			continue;

		if (!match_oui0((unsigned char *)n->oui0, as->hwaddr, n->ouis))
			continue;

		if (memcmp(sta, as->hwaddr, 6) == 0)
			continue;

		dbg("Disconnecting redundant repeated client " MACFMT "\n",
				MAC2STR(as->hwaddr));

		invoke_disconnect_sta_by_bssid(c, as->bssid, as->hwaddr);
		invoke_disconnect_sta_by_bssid(c, as->old_bssid, as->hwaddr);
	}
}

static int parse_sta_event(struct controller *c, const char *evtype,
						struct blob_attr *msg)
{
	struct netif_fhbss *p;
	char macaddr[18] = {0}, action[64] = {0}, vif[16] = {0};
	unsigned char sta_macaddr[6], bssid[6] = {0};
	bool found = false;
	struct active_sta *as;
	uint32_t lowrssi = 0;
	int rv;
	enum {
		MACADDR,
		ACTION,
		VIF,
		LOWRSSI,
		__POLICY_MAX
	};
	const struct blobmsg_policy sta_policy[__POLICY_MAX] = {
		[MACADDR] = {.name = "macaddr", .type = BLOBMSG_TYPE_STRING},
		[ACTION] = {.name = "action", .type = BLOBMSG_TYPE_STRING},
		[VIF] = {.name = "vif", .type = BLOBMSG_TYPE_STRING},
		[LOWRSSI] = {.name = "lowrssi", .type = BLOBMSG_TYPE_INT32},
	};
	struct blob_attr *tb[__POLICY_MAX];

	char sta_node[128] = {0};
	char *sta_p;
	struct in_addr ipn;
	struct node *n;

	sta_p = strstr(evtype, "wifi.sta");
	snprintf(sta_node, sta_p - evtype, "%s", evtype);
	if (strlen(sta_node) && !inet_aton(sta_node, &ipn)) {
		warn("Invalid 'wifi.sta' prefix: %s\n", sta_node);
		return -1;
	}

	if (strlen(sta_node))
		n = find_node_by_ip(c, sta_node);
	else
		n = find_node_by_ip(c, "127.0.0.1");

	rv = blobmsg_parse(sta_policy, __POLICY_MAX, tb,
			blobmsg_data(msg), blobmsg_len(msg));
	if (rv < 0) {
		cntlr_dbg("Error parsing wifi.sta events!\n");
		goto fail;
	}

	if (!tb[MACADDR] || !tb[VIF] || !tb[ACTION])
		goto fail;

	strncpy(macaddr, blobmsg_get_string(tb[MACADDR]), 17);
	if (hwaddr_aton(macaddr, sta_macaddr) == NULL)
		goto fail;

	strncpy(vif, blobmsg_get_string(tb[VIF]), 15);

	if (tb[LOWRSSI]) {
		lowrssi = blobmsg_get_u32(tb[LOWRSSI]);
		if (lowrssi != 0) {
			cntlr_dbg("Received wifi.sta 'lowrssi' (%d) event " \
					"+++++++++++++>\n", lowrssi);
			return cntlr_handle_sta_rssi_low(c, n, sta_macaddr, lowrssi);
		}
	}

	strncpy(action, blobmsg_get_string(tb[ACTION]), 63);

	/* At this instant when the associated STA is checked to see if it is
	 * backhaul wifi-iface of a node, the nodelist does not include this
	 * STA.
	 * We check again for pure STA from wifi.agent objects' appear/disappear
	 * event handler.
	 */
	if (!endpoint_hwaddr(c, sta_macaddr)) {
		cntlr_dbg("STA %s is NOT a pure STA!\n", macaddr);
		goto fail;
	}

	/* TODO: wifi.sta events should contain bssid too */
	list_for_each_entry(p, &n->iflist, list) {
		if (strncmp(p->ifname, vif, 16) != 0)
			continue;

		memcpy(bssid, p->bssid, 6);
		found = true;
		break;
	}

	if (!found)
		goto fail;

	if (strncmp(action, "disassociated", strlen("disassociated") + 1) == 0) {
		as = as_lookup(c->as_table, sta_macaddr);

		/* if disassociate event came from most recently recorded bssid,
		 * we must have missed the disassociate event from previous one
		 */
		if (as && memcmp(as->bssid, bssid, 6) == 0)
			invoke_disconnect_sta_by_bssid(n->cntlr,
							as->old_bssid,
							as->hwaddr);

		as_clean_entry(c->as_table, sta_macaddr, bssid);
	} else if (strncmp(action, "associated", strlen("associated") + 1) == 0) {
		as = as_lookup(c->as_table, sta_macaddr);
		if (as) {
			if (memcmp(as->bssid, bssid, 6) != 0) {
				struct node *prev_ap;
				char hwaddrstr[12 + 1] = {0};

				/* if entry already exists, it must not have
				 * receieved dissassoc event;
				 * issue disconnect_sta to previous bssid.
				 */
				invoke_disconnect_sta_by_bssid(n->cntlr,
								as->bssid,
								as->hwaddr);

				/* flush from arp table of previous bssid */
				prev_ap = get_node_by_bssid(c, as->bssid);
				if (!prev_ap)
					return -1;

				btostr(as->hwaddr, sizeof(as->hwaddr),
						hwaddrstr);
			}
		}

		/* disconnect redundant repeated addresses if real mac is seen
		 * locally, or a repeated mac is seen
		 */
		/* TODO TODO: check this logic.. again */
		if ((sta_macaddr[1] != n->mobility_domain[0] ||
			sta_macaddr[2] != n->mobility_domain[1]) ||
			(sta_macaddr[1] == n->mobility_domain[0] &&
			 sta_macaddr[2] == n->mobility_domain[1] &&
			match_oui0(n->oui0, sta_macaddr, n->ouis))) {

			disconnect_repeated_macs(c, n, sta_macaddr);
		}

		as = as_insert(c->as_table, sta_macaddr, bssid);

		/* instruct agents to start monitoring this sta */
		cntlr_dbg("Start Monitoring for STA " MACFMT " ....\n",
				MAC2STR(sta_macaddr));
		cntlr_monitor_sta(c, sta_macaddr, n);
	}

	as_print(c->as_table);
	return 0;
fail:
	return -1;
}

static void node_event_handler(struct ubus_context *ctx,
					struct ubus_event_handler *ev,
					const char *type,
					struct blob_attr *msg)
{
	struct node *n = container_of(ev, struct node, evh);
	char *str;

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	cntlr_dbg("[ &node = %p ] Received event = '%s': %s\n", n, type, str);
	if (strstr(type, "wifi.sta"))
		parse_sta_event(n->cntlr, type, msg);

	/* handle other events as needed */

	free(str);
}
#endif

struct netif_iface *cntlr_radio_add_interface(struct controller *c,
					      struct netif_radio *r,
					      uint8_t *hwaddr)
{
	struct netif_iface *n;

	n = find_interface_by_mac(c, r, hwaddr);
	if (n) {
		n->active = true;
		n->band = r->band;
		return n;
	}

	n = calloc(1, sizeof(*n));
	if (!n)
		return NULL;

	n->metrics = calloc(12, sizeof(*n->metrics)); /* TODO: dummy value to 1 */
	if (!n->metrics) {
		trace("|%s:%d| Alloc error\n");
		free(n);
		return NULL;
	}

	memcpy(n->bssid, hwaddr, 6);
	n->type = NETIF_FHBSS;
	n->active = true;
	list_add(&n->list, &r->iflist);
	n->agent = r->agent;
	n->band = r->band;
	return n;
}

struct netif_radio *cntlr_node_add_radio(struct controller *c, struct node *n,
		uint8_t *radio)
{
	struct netif_radio *r;

	r = find_radio_by_node(c, n, radio);
	// trace("-------------------> %s : raadio added "MACFMT"\n", MAC2STR(r->hwaddr));
	if (r)
		return r;

	r = calloc(1, sizeof(*r));
	if (!r)
		return NULL;

	memcpy(r->macaddr, radio, 6);
	INIT_LIST_HEAD(&r->iflist);
	list_add(&r->list, &n->radiolist);
	r->agent = n;

	return r;
}

static int get_op_class_bw(int op_class)
{
	switch (op_class) {
	case 115:
	case 118:
	case 121:
	case 125:
	case 81:
	case 82:
	case 124:
		return 20;
	case 116:
	case 119:
	case 122:
	case 117:
	case 120:
	case 123:
	case 83:
	case 84:
	case 126:
	case 127:
		return 40;
	case 128:
	case 130:
		return 80;
	case 129:
		return 160;
	default:
		return 20;
	}
}

static enum wifi_freqband get_op_class_band(int opclass)
{
	switch (opclass) {
	case 81:
	case 82:
	case 83:
	case 84:
		return BAND_2;
	case 115:
	case 116:
	case 117:
	case 118:
	case 119:
	case 120:
	case 121:
	case 122:
	case 123:
	case 124:
	case 125:
	case 126:
	case 127:
	case 128:
	case 129:
	case 130:
		return BAND_5;
	default:
		return BAND_UNKNOWN;
	}
}

static struct opclass_entry *cntlr_radio_opclass_find_entry(struct opclass *opclass, uint8_t id)
{
	struct opclass_entry *entry;
	int i;

	for (i = 0; i < opclass->opclass_entry_num; i++) {
		entry = &opclass->opclass_entry[i];
		if (entry->opclass != id)
			continue;

		return entry;
	}

	return NULL;
}

static struct opclass_entry *cntlr_radio_opclass_new_entry(struct opclass *opclass)
{
	struct opclass_entry *entry;

	if (opclass->opclass_entry_num >= ARRAY_SIZE(opclass->opclass_entry))
		return NULL;

	entry = &opclass->opclass_entry[opclass->opclass_entry_num];
	opclass->opclass_entry_num++;
	opclass->opclass_entry_time = time(NULL);

	return entry;
}

static int cntlr_radio_opclass_add_channel(struct opclass_entry *entry, uint8_t channel, uint8_t preference)
{
	int i;

	/* First check if we already have this */
	for (i = 0; i < entry->channels_num; i++) {
		if (entry->channels[i].channel == channel) {
			entry->channels[i].preference = preference;
			return 0;
		}
	}

	/* Add new channel */
	if (entry->channels_num >= ARRAY_SIZE(entry->channels))
		return -1;

	entry->channels[entry->channels_num].channel = channel;
	entry->channels[entry->channels_num].preference = preference;
	entry->channels_num++;

	return 0;
}

static void _cntlr_radio_opclass_reset(struct opclass *opclass)
{
	opclass->opclass_entry_num = 0;
	memset(opclass->opclass_entry, 0, sizeof(opclass->opclass_entry));
}

static void _cntlr_radio_opclass_dump(struct opclass *opclass)
{
	struct opclass_entry *entry;
	int i, j;

	dbg(">>> opclass num: %d\n", opclass->opclass_entry_num);
	for (i = 0; i < opclass->opclass_entry_num; i++) {
		entry = &opclass->opclass_entry[i];
		dbg("opclass: %u\n", entry->opclass);
		for (j = 0; j < entry->channels_num; j++) {
			dbg("\tchan %u pref %u reason %u\n",
			    entry->channels[j].channel,
			    (entry->channels[j].preference & CHANNEL_PREF_MASK) >> 4,
			    entry->channels[j].preference & CHANNEL_PREF_REASON);
		}
	}
	dbg("<<<\n");
}

void cntlr_radio_opclass_reset(struct netif_radio *radio)
{
	_cntlr_radio_opclass_reset(&radio->opclass);
}

int cntlr_radio_opclass_add(struct netif_radio *radio, uint8_t opclass,
			    uint8_t channel, uint8_t preference)
{
	struct opclass_entry *entry;

	entry = cntlr_radio_opclass_find_entry(&radio->opclass, opclass);
	if (!entry)
		entry = cntlr_radio_opclass_new_entry(&radio->opclass);
	if (!entry)
		return -1;

	entry->opclass = opclass;
	entry->bw = get_op_class_bw(opclass);
	return cntlr_radio_opclass_add_channel(entry, channel, preference);
}

void cntlr_radio_opclass_dump(struct netif_radio *radio)
{
	_cntlr_radio_opclass_dump(&radio->opclass);
}

void cntlr_radio_cur_opclass_reset(struct netif_radio *radio)
{
	_cntlr_radio_opclass_reset(&radio->cur_opclass);
}

int cntlr_radio_cur_opclass_add(struct netif_radio *radio, uint8_t opclass,
				uint8_t channel, uint8_t txpower)
{
	struct opclass_entry *entry;

	entry = cntlr_radio_opclass_find_entry(&radio->cur_opclass, opclass);
	if (!entry)
		entry = cntlr_radio_opclass_new_entry(&radio->cur_opclass);
	if (!entry)
		return -1;

	entry->opclass = opclass;
	entry->bw = get_op_class_bw(opclass);
	entry->txpower = txpower;
	return cntlr_radio_opclass_add_channel(entry, channel, 15 << 4);
}

void cntlr_radio_cur_opclass_dump(struct netif_radio *radio)
{
	_cntlr_radio_opclass_dump(&radio->cur_opclass);
}

static bool cntlr_radio_opclass_expired(struct netif_radio *radio)
{
	time_t age;

	if (!radio->opclass.opclass_entry_num)
		return true;

	age = time(NULL) - radio->opclass.opclass_entry_time;
	if (((unsigned int) age) > 120)
		return true;

	return false;
}

bool cntlr_node_opclass_expired(struct node *node)
{
	struct netif_radio *radio = NULL;
	bool expired = false;

	list_for_each_entry(radio, &node->radiolist, list) {
		expired |= cntlr_radio_opclass_expired(radio);
	}

	return expired;
}

static enum wifi_freqband cntlr_radio_opclass_get_band(struct opclass *opclass)
{
	enum wifi_freqband band;
	enum wifi_freqband ret = 0;
	int i;

	for (i = 0; i < opclass->opclass_entry_num; i++) {
		band = get_op_class_band(opclass->opclass_entry[i].opclass);
		if (band == BAND_UNKNOWN)
			continue;
		ret |= band;
	}

	return ret;
}

void cntlr_radio_update_band(struct netif_radio *radio)
{
	radio->band = cntlr_radio_opclass_get_band(&radio->cur_opclass);
}

struct node *alloc_node_init(struct controller *c, uint8_t *hwaddr)
{
	struct node *n;
	struct cmdu_buff *cmdu;

	n = find_node_by_mac(c, hwaddr);
	if (n)
		return n;

	n = calloc(1, sizeof(struct node));
	if (!n) {
		warn("OOM: node malloc failed!\n");
		return NULL;
	}

	n->cntlr = c;
	n->depth = -1;
	n->scan_supported = true;
	n->np = NULL; //c->cfg.apolicy;
	memcpy(n->alid, hwaddr, 6);

	//INIT_LIST_HEAD(&n->stalist);
	INIT_LIST_HEAD(&n->radiolist);
	list_add(&n->list, &c->nodelist);
	c->num_nodes++;

	dbg("%s %d --------------------------- " MACFMT "\n", __func__, __LINE__, MAC2STR(hwaddr));

	cmdu = cntlr_gen_bk_caps_query(c, n->alid);
	if (!cmdu)
		return n;

	send_cmdu(c, cmdu);
	cmdu_free(cmdu);

	cmdu = cntlr_gen_ap_capability_query(c, n->alid);
	if (!cmdu)
		return n;

	send_cmdu(c, cmdu);
	cmdu_free(cmdu);

	return n;
}

void node_clean_stalist(struct controller *c)
{
	struct sta *s = NULL, *tmp;

	list_for_each_entry_safe(s, tmp, &c->stalist, list) {
		list_del(&s->list);
		free(s);
	}
}

void node_clean_linklist(struct controller *c)
{
        struct netif_link *l = NULL, *tmp;

        list_for_each_entry_safe(l, tmp, &c->linklist, list) {
		free(l->metrics);
                list_del(&l->list);
                free(l);
        }
}


static void radio_clean_iflist(struct netif_radio *r)
{
	struct netif_iface *ni = NULL, *tmp;

	list_for_each_entry_safe(ni, tmp, &r->iflist, list) {
		list_del(&ni->list);
		free(ni->metrics);
		free(ni);
	}
}
static void node_clean_radiolist(struct node *n)
{
	struct netif_radio *r = NULL, *tmp;

	list_for_each_entry_safe(r, tmp, &n->radiolist, list) {
		list_del(&r->list);
		radio_clean_iflist(r);
		free(r);
	}
}

static void cntlr_clean_nodelist(struct controller *c)
{
	struct node *n = NULL, *tmp;

	list_for_each_entry_safe(n, tmp, &c->nodelist, list) {
		node_clean_stalist(c);
		node_clean_linklist(c);
		node_clean_radiolist(n);
		list_del(&n->list);
		free(n);
	}
}

void free_bcn_metrics(struct controller *c, struct sta *s)
{
	struct bcn_metrics *b = NULL, *tmp;

	list_for_each_entry_safe(b, tmp, &s->bcnlist, list) {
		list_del(&b->list);
		free(b);
		s->num_bcn_metrics--;
	}
}

void cntlr_mark_old_bcn_metrics(struct controller *c, struct sta *s)
{
	struct bcn_metrics *b = NULL, *tmp;

	list_for_each_entry_safe(b, tmp, &s->bcnlist, list) {
		/* TODO: only most recent measurement marked as fresh */
		b->stale = true;
	}
}

void free_usta_metrics(struct controller *c, struct sta *s)
{
	struct una_sta_metrics *u = NULL, *tmp;

	list_for_each_entry_safe(u, tmp, &s->unassoclist, list) {
		list_del(&u->list);
		free(u);
	}
}

struct netif_link *alloc_link_init(struct controller *c,
		uint8_t *upstream, uint8_t *downstream)
{
	struct netif_link *l;

	l = find_link_by_mac(c, upstream, downstream);
	if (l)
		return l;

	l = calloc(1, sizeof(struct netif_link));
	if (!l)
		return NULL;

	l->metrics = calloc(1, sizeof(struct link_metrics));
	if (!l->metrics)
		goto out;

	l->upstream = find_interface_by_mac_nor(c, upstream);
	if (!l->upstream)
		goto out_metrics;

	l->downstream = find_interface_by_mac_nor(c, downstream);
	if (!l->downstream)
		goto out_metrics;

	trace("Adding link | " MACFMT " <---> " MACFMT " |\n",
	      MAC2STR(l->upstream->bssid), MAC2STR(l->downstream->bssid));
	list_add(&l->list, &c->linklist);

	return l;

out_metrics:
	free(l->metrics);
out:
	free(l);
	return NULL;
}

void free_watchnode_cleanup(struct controller *c, struct watchnode *wn)
{
	uloop_timeout_cancel(&wn->scan_timer);
	uloop_timeout_cancel(&wn->scanres_timer);
	list_del(&wn->list);
	free(wn);
}

static void cntlr_radar_exit(struct uloop_timeout *t)
{
	/*TODO: before change channel due to radar, save old chandef.
	 * Restore that chandef upon exit from radar nop.
	 */
}

#if 0
static int cntlr_handle_radar_alert(struct controller *c, struct watchnode *wn)
{
	return 0;
}

static void forall_node_exclude_steering(struct controller *c)
{
}

static void forall_node_trigger_scan(struct controller *c)
{
}

static void perform_tasks_topology_init(struct controller *c)
{
	//forall_node_get_fhinfo(c);
	//forall_node_get_assoclist(c);
}

static void perform_tasks_topology_change(struct controller *c)
{
	forall_node_exclude_steering(c);
	forall_node_trigger_scan(c);
	forall_node_get_fhinfo(c);
}
#endif


static void cntlr_ieee1905_cmdu_event_handler(void *cntlr,
		struct blob_attr *msg)
{
	static const struct blobmsg_policy cmdu_attrs[6] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT16 },
		[1] = { .name = "mid", .type = BLOBMSG_TYPE_INT16 },
		[2] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "source", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "origin", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "cmdu", .type = BLOBMSG_TYPE_STRING },
	};
	struct controller *c = (struct controller *)cntlr;
	char in_ifname[16] = {0};
	struct blob_attr *tb[6];
	char src[18] = { 0 }, src_origin[18] = { 0 };
	uint8_t *tlv = NULL;
	char *tlvstr = NULL;
	uint16_t type;
	uint8_t srcmac[6], origin[6];
	uint16_t mid = 0;
	int len = 0;
	sigset_t waiting_mask;

	sigpending(&waiting_mask);
	if (sigismember(&waiting_mask, SIGINT) ||
			sigismember(&waiting_mask, SIGTERM))
		return;

	blobmsg_parse(cmdu_attrs, 6, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	if (tb[0]) {
		int t;

		t = blobmsg_get_u16(tb[0]);
		if (t < 0)
			return;

		type = (uint16_t)t;
		if (!is_cmdu_for_us(c, type))
			return;
	}

	if (tb[1])
		mid = (uint16_t)blobmsg_get_u16(tb[1]);


	if (tb[2])
		strncpy(in_ifname, blobmsg_data(tb[2]), 15);

	if (tb[3]) {
		strncpy(src, blobmsg_data(tb[3]), 17);
		hwaddr_aton(src, srcmac);
	}

	if (tb[4]) {
		strncpy(src_origin, blobmsg_data(tb[4]), 17);
		hwaddr_aton(src_origin, origin);
	}


	if (tb[5]) {
		len = blobmsg_data_len(tb[5]) - 16;

		tlvstr = calloc(1, len + 1);

		if (!tlvstr)
			return;

		strncpy(tlvstr, (blobmsg_data(tb[5]) + 16), len);
		len = (len - 1) / 2;
		tlv = calloc(1, len);
		if (!tlv) {
			free(tlvstr);
			return;
		}

		strtob(tlvstr, len, tlv);
		free(tlvstr);
	}
	cntlr_handle_map_event(c, type, mid, in_ifname, srcmac, origin, tlv, len);

	if (tlv)
		free(tlv);
}

#if 0
static int cntlr_register_events(struct controller *c)
{
	cntlr_dbg("Setting up bus event handlers...\n");

	return 0;
}
#endif

static int ubus_call_object(struct controller *c, uint32_t obj,
		const char *method,
		void (*response_cb)(struct ubus_request *, int, struct blob_attr *),
		void *priv)
{
	struct blob_buf bb = {};
	int ret;

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(c->ubus_ctx, obj, method, bb.head,
				response_cb, priv, 2 * 1000);
	if (ret) {
		err("Failed to get '%s' (ret = %d)\n", method, ret);
		blob_buf_free(&bb);
		return -1;
	}

	blob_buf_free(&bb);
	return 0;
}

#if 0
static uint32_t ubus_get_object(struct ubus_context *ctx, const char *name)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(ctx, name, &id);
	if (status != UBUS_STATUS_OK) {
		err("object '%s' not present!\n", name);
		return -1;
	}

	return id;
}

/* This function will also be called from topology change event */
static int get_topology(struct controller *c)
{
	struct blob_buf bb = {};
	int ret;

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(c->ubus_ctx, c->topology, "tree", bb.head,
			build_topology_tree, c, 2 * 1000);
	if (ret) {
		cntlr_dbg("Failed to get topology tree (ret = %d)\n", ret);
		blob_buf_free(&bb);
		return -1;
	}

	blob_buf_free(&bb);
	return 0;
}

static void uobjx_lookup(struct ubus_context *ctx,
				struct ubus_object_data *obj, void *priv)
{
	struct controller *c = priv;
	const char *token = "wifi.agent";
	char *p;

	p = strstr(obj->path, token);
	if (p != NULL) {
		char node[128] = {0};

		snprintf(node, p - obj->path, "%s", obj->path);
		topology_add_node(c, node);
	}
}

static int enumerate_topology_indirect(struct controller *c)
{
	const char *prefix = NULL;

	ubus_lookup(c->ubus_ctx, prefix, uobjx_lookup, c);

	return 0;
}

static uint32_t ubus_get_topology_object(struct ubus_context *ctx)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(ctx, "topology", &id);
	if (status != UBUS_STATUS_OK) {
		warn("object 'topology' not present!\n");
		return OBJECT_INVALID;
	}

	return id;
}
#endif


static void cntlr_query_nodes(struct uloop_timeout *t)
{
	struct controller *c = container_of(t, struct controller, query_nodes);
	struct node *n;

	list_for_each_entry(n, &c->nodelist, list) {
		struct cmdu_buff *cmdu;

		cmdu = cntlr_gen_bk_caps_query(c, n->alid);
		if (cmdu) {
			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
		}

		cmdu = cntlr_gen_ap_capability_query(c, n->alid);
		if (cmdu) {
			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
		}

		cmdu = cntlr_gen_topology_query(c, n->alid);
		if (cmdu) {
			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
		}
	}

	uloop_timeout_set(&c->query_nodes, 60 * 1000);
}

#define RADIO_MAX 8
bool cntlr_check_config_diff(struct controller *c, uint8_t diff)
{
	bool reloaded = false;

	if (diff & CONFIG_DIFF_CREDENTIALS || diff & CONFIG_DIFF_VLAN) {
		struct cmdu_buff *cmdu;
		uint8_t origin[6] = {0x01, 0x80, 0xc2, 0x00, 0x00, 0x13};

		trace("Config changed, triggering renew!\n");
		cmdu = cntlr_gen_ap_autoconfig_renew(c, origin);
		if (cmdu) {
			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
			reloaded = true;
		}
	} else if (diff & (CONFIG_DIFF_AGENT_POLICY | CONFIG_DIFF_AGENT_POLICY_CNT)) {
		struct node_policy *p;
		struct node *n = NULL;


		/* TODO/CLEANUP:
		 * as of now, no information is being stored about
		 * the specific agent's radios & BSS,
		 * Also few information i.e. exclude stalist,
		 * rcpi/util threshold is sent along with
		 * 17.2.11 & 17.2.12 tlv.
		 * So for now, dummy radio id & bss id is being used for this
		 * purpose.
		 * ((later on radio id & bss id info for specific agent
		 * will be stored using in Topology Response CMDU.))
		 */

		trace("agent policy config changed\n");

		/* send the policy config cmdu to the marked agent */
		list_for_each_entry(n, &c->nodelist, list) {
			struct cmdu_buff *cmdu;
			int num_bk = 1;
			uint8_t bk_id[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
			struct radio_policy *rp = NULL;
			uint8_t radio_id[RADIO_MAX * 6] = {0};
			int num_radio = 0;

			if ((diff & CONFIG_DIFF_AGENT_POLICY) && !n->np->is_policy_diff)
				continue;

			list_for_each_entry(rp, &n->np->radiolist, list) {
				memcpy(&radio_id[num_radio * 6],
					rp->macaddr,
					6);
				num_radio++;
			}

			cmdu = cntlr_gen_policy_config_req(c, n->alid,
					n->np, num_radio, radio_id, num_bk,
					bk_id);
			if (cmdu) {
				send_cmdu(c, cmdu);
				cmdu_free(cmdu);
				reloaded = true;
			}
		}

		/* reset is_policy_diff to false; */
		list_for_each_entry(p, &c->cfg.nodelist, list) {
			p->is_policy_diff = false;
		}
	}

	return reloaded;
}

#ifdef CONTROLLER_SYNC_DYNAMIC_CNTLR_CONFIG
int cntlr_sync_dyn_controller_config(struct controller *c, uint8_t *agent)
{
	struct node *n;
	uint8_t proto = 0xab;
	struct cmdu_buff *cmdu;

	if (agent && !hwaddr_is_zero(agent)) {
		cmdu = cntlr_gen_higher_layer_data(c, agent, proto, NULL, 0);
		if (!cmdu)
			return -1;

		send_cmdu(c, cmdu);
		cmdu_free(cmdu);
	} else {
		list_for_each_entry(n, &c->nodelist, list) {
			if (hwaddr_equal(c->almac, n->alid))
				continue; /* skip locally running agent */

			cmdu = cntlr_gen_higher_layer_data(c, n->alid, proto, NULL, 0);
			if (!cmdu)
				return -1;

			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
		}
	}

	return 0;
}
#endif

bool cntlr_resync_config(struct controller *c, bool reload)
{
	uint8_t diff;
	struct node_policy *np = NULL;

	diff = cntlr_config_reload(&c->cfg);

	list_for_each_entry(np, &c->cfg.nodelist, list) {
		struct node *n;

		n = find_node_by_mac(c, np->agent_id);
		if (n)
			n->np = np;
	}

	if (reload)
		cntlr_check_config_diff(c, diff);

#ifdef CONTROLLER_SYNC_DYNAMIC_CNTLR_CONFIG
	/* in dyn-controller mode, sync controller's config in network */
	if (diff)
		cntlr_sync_dyn_controller_config(c, NULL);
#endif

	return !!diff;
}

static void cntlr_signal_periodic_run(struct uloop_timeout *t)
{
	struct controller *c = container_of(t, struct controller, signal_handler);
	sigset_t waiting_mask;

	sigpending(&waiting_mask);

	if (sigismember(&waiting_mask, SIGHUP)) {
		dbg("|%s:%d| Received SIGHUP, reload config\n", __func__, __LINE__);
		signal(SIGHUP, SIG_IGN);

		cntlr_resync_config(c, true);
	}

	uloop_timeout_set(&c->signal_handler, 1 * 1000);
}

int update_txlink_metric_data(struct controller *c, struct tlv_tx_linkmetric *txl, int len)
{
	int i = 0;
	int size = 0;
	struct tx_link_info *txlinfo;
	struct netif_link *txlink;
	struct link_metrics *metrics;

	/* For each tx link in the mesg */
	size = len - (sizeof(struct tlv_tx_linkmetric));
	size = size / (sizeof(struct tx_link_info));
	for (i = 0; i < size; i++) {
		txlinfo = (struct tx_link_info *)&txl->link[i];
		/* Find or alloc the backhaul link to the controller */
		txlink = alloc_link_init(c, txlinfo->local_macaddr, txlinfo->neighbor_macaddr);
		if (!txlink) {
			trace("No link!\n");
			return -1;
		}
		metrics = txlink->metrics;
		metrics->l = txlink;
		memcpy(metrics->l->upstream->bssid, txlinfo->local_macaddr, 6);
		memcpy(metrics->l->downstream->bssid, txlinfo->neighbor_macaddr, 6);
		metrics->type = BUF_GET_BE16(txlinfo->mediatype);
		metrics->bridge = txlinfo->has_bridge;
		metrics->packet_tx_error = BUF_GET_BE32(txlinfo->errors);
		metrics->packet_trans = BUF_GET_BE32(txlinfo->packets);
		metrics->thp = BUF_GET_BE16(txlinfo->max_throughput);
		metrics->link_av = BUF_GET_BE16(txlinfo->availability);
		metrics->phy_rate = BUF_GET_BE16(txlinfo->phyrate);
	}
	c->num_tx_links += size;
	return 0;
}

int update_rxlink_metric_data(struct controller *c, struct tlv_rx_linkmetric *rxl, int len)
{
	int i = 0;
	int size = 0;
	struct rx_link_info *rxlinfo;
	struct netif_link *rxlink;
	struct link_metrics *metrics;

	/* For each rx link in the mesg */
	size = len - (sizeof(struct tlv_rx_linkmetric));
	size = size / (sizeof(struct rx_link_info));
	for (i = 0; i < size; i++) {
		rxlinfo = (struct rx_link_info *)&rxl->link[i];
		/* Find or alloc the backhaul link to the controller */
		rxlink = alloc_link_init(c, rxlinfo->local_macaddr, rxlinfo->neighbor_macaddr);
		if (!rxlink) {
			trace("No link!\n");
			return 0;
		}
		metrics = rxlink->metrics;
		metrics->l = rxlink;
		memcpy(metrics->l->upstream->bssid, rxlinfo->local_macaddr, 6);
		memcpy(metrics->l->downstream->bssid, rxlinfo->neighbor_macaddr, 6);
		metrics->type = BUF_GET_BE16(rxlinfo->mediatype);
		metrics->packet_rec = BUF_GET_BE32(rxlinfo->packets);
		metrics->packet_rx_error = BUF_GET_BE32(rxlinfo->errors);
		metrics->rssi = rxlinfo->rssi;
	}
	c->num_rx_links += size;
	return 0;
}

int handle_link_metrics_response(struct controller *c, struct cmdu_buff *cstruct)
{
	trace("%s: --->\n", __func__);
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_TRANSMITTER_LINK_METRIC, .present = TLV_PRESENT_MORE},
		[1] = { .type = TLV_TYPE_RECEIVER_LINK_METRIC, .present = TLV_PRESENT_MORE},
	};
	struct tlv *tv[2][16];
	int num = 0;
	int ret;

	if (cstruct == NULL || c == NULL)
		return -1;

	/* Reset number of links */
	c->num_tx_links = 0;
	c->num_rx_links = 0;

	/* Parsing 1905 Link Metrics TLV */
	cmdu_parse_tlvs(cstruct, tv, a_policy, 2);

	/* Storing 1905 Link Metrics TLV */
	while (tv[0][num]) {
		struct tlv_tx_linkmetric *txl =
			(struct tlv_tx_linkmetric *)tv[0][num]->data;

		if (hwaddr_is_zero(txl->aladdr) ||
			hwaddr_is_zero(txl->neighbor_aladdr)) {
			trace("%s: Discard Tx-link response\n", __func__);
			return -1;
		}
		/* Storing tx 1905 Link Metric data */
		ret = update_txlink_metric_data(c, txl, BUF_GET_BE16(tv[0][num]->len));
		if (ret) {
			trace("|update_txlink_metric_data| Link not found!\n");
			return -1;
		}
		num++;
	}

	num = 0;
	while (tv[1][num]) {
		struct tlv_rx_linkmetric *rxl =
			(struct tlv_rx_linkmetric *)tv[1][num]->data;

		if (hwaddr_is_zero(rxl->aladdr) ||
			hwaddr_is_zero(rxl->neighbor_aladdr)) {
			trace("%s: Discard Tx-link response\n", __func__);
			return -1;
		}
		/* Storing rx 1905 Link Metric data */
		ret = update_rxlink_metric_data(c, rxl, BUF_GET_BE16(tv[0][num]->len));
		if (ret) {
			trace("|update_txlink_metric_data| Link not found!\n");
			return -1;
		}
		num++;
	}

	return 0;
}

static void send_buildcmdu_cb(struct ubus_request *req,
		int type, struct blob_attr *msg)
{
	char tlv_str[500] = {0};
	int len, b_len;
	uint16_t msg_type;
	uint8_t *tlv = NULL;
	struct cmdu_buff *cmdu = NULL;
	struct controller *c;
	uint16_t mid = 0;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy cb_attr[2] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "data", .type = BLOBMSG_TYPE_STRING },
	};

	blobmsg_parse(cb_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[1] || !tb[0])
		return;

	c  = (struct controller *)req->priv;
	msg_type = (uint16_t)blobmsg_get_u32(tb[0]);
	strncpy(tlv_str, blobmsg_data(tb[1]), sizeof(tlv_str) - 1);
	len = strlen(tlv_str);
	b_len = (len/2);
	tlv = (uint8_t *) malloc((b_len) * sizeof(uint8_t));
	if (tlv == NULL) {
		err("No Memory\n");
		return;
	}
	strtob(tlv_str, b_len, tlv);

	cmdu = cmdu_alloc_custom(msg_type, &mid, NULL, c->almac, tlv, b_len);
	if (!cmdu)
		goto out;

	handle_link_metrics_response(c, cmdu);

	cmdu_free(cmdu);
out:
	free(tlv);
	return;
}

int send_buildcmdu(struct controller *c, uint16_t msg_type)
{
	struct blob_buf b = { 0 };
	int ret = 0;
	uint32_t id;

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	blobmsg_add_u32(&b, "type", (uint32_t)msg_type);

	if (!c->ubus_ctx)
		trace("I think that the ubus_ctx is NULL\n");
	if (ubus_lookup_id(c->ubus_ctx, "ieee1905", &id)) {
		dbg("[%s:%d] not present i1905", __func__, __LINE__);
		goto out;
	}

	trace("\t\t%s: %d\n", __func__, __LINE__);

	ret = ubus_invoke(c->ubus_ctx, id, "buildcmdu",
			b.head, send_buildcmdu_cb, c, 5000);
	if (ret) {
		dbg("[%s:%d] ubus call failed for |i1905 buildcmdu|",
			 __func__, __LINE__);
		goto out;
	}
out:
	blob_buf_free(&b);
	return ret;
}

static void combined_link_metric_periodic_collection(struct controller *c)
{
	trace("%s: --->\n", __func__);
	struct cmdu_buff *cmdu;
	struct node *p;
	struct netif_radio *radio;
	struct netif_iface *bss;
	uint8_t *bsslist = NULL, *new_bsslist = NULL;
	uint8_t *radiolist = NULL, *new_radiolist = NULL;
	int num_bss, num_radio;
	int radio_index, bss_index;
	uint8_t hwaddr[6];

	/* AP metrics query for each agent */
	/* For each agent */
	list_for_each_entry(p, &c->nodelist, list) {
		num_radio = 0;
		num_bss = 0;
		memcpy(hwaddr, p->alid, 6);
		/* For each radio */
		list_for_each_entry(radio, &p->radiolist, list) {
			/* Building a radiolist of all radios */
			new_radiolist = (uint8_t *)realloc(radiolist,
							6 * (num_radio + 1) * sizeof(uint8_t));

			if (!new_radiolist) {
				trace("realloc of radiolist failed\n");
				goto error;
			}

			radiolist = new_radiolist;
			num_radio++;
			radio_index = (num_radio - 1) * 6;
			memcpy(radiolist + radio_index, radio->macaddr, 6);

			/* For each bss in radio */
			list_for_each_entry(bss, &radio->iflist, list) {
				if (bss->type) /*if bss is bsta */
					continue;

				/* Building a bsslist of all BSS */
				new_bsslist = (uint8_t *)realloc(bsslist,
							6 * (num_bss + 1) * sizeof(uint8_t));

				if (!new_bsslist) {
					trace("realloc of bsslist failed\n");
					goto error;
				}

				bsslist = new_bsslist;
				num_bss++;
				bss_index = (num_bss - 1) * 6;
				memcpy(bsslist + bss_index, bss->bssid, 6);
			}
		}
		cmdu = cntlr_gen_ap_metrics_query(c, hwaddr, num_bss, bsslist, num_radio, radiolist);
		if (!cmdu) {
			trace("cmdu_gen failed!\n");
			goto error;
		}

		send_cmdu(c, cmdu);

		/* 1905 Link metric query */
		send_buildcmdu(c, CMDU_TYPE_LINK_METRIC_RESPONSE);
		cmdu_free(cmdu);
	}
error:
	if (radiolist)
		free(radiolist);
	if (bsslist)
		free(bsslist);
}

static void cntlr_periodic_run(struct uloop_timeout *t)
{
	struct controller *c = container_of(t, struct controller, heartbeat);

	cntlr_log_nodes(c);

	//forall_node_get_fhinfo(c);   /* replaced from per-node refresh bss */

	forall_node_get_sta_metrics(c);

	/* TODO: */
	//forall_node_get_usta_metrics(c);

	/* TODO: update only when a node is added or removed */
	//forall_node_update_neighbors(c);

	/* Call AP metrics query and 1905 link metrics query for data collection */
	combined_link_metric_periodic_collection(c);

	uloop_timeout_set(&c->heartbeat, 10 * 1000);
}

static void cntlr_acs_run(struct uloop_timeout *t)
{
	struct controller *c = container_of(t, struct controller, acs);
	bool skip_dfs = false;

	/* Run ACS recalc here */
	dbg("acs timeout - run recalc\n");
	cntlr_acs_recalc(c, skip_dfs);

	if (c->cfg.acs_timeout)
		uloop_timeout_set(&c->acs, c->cfg.acs_timeout * 1000);
}

static void cntlr_dfs_cleanup_run(struct uloop_timeout *t)
{
	struct controller *c = container_of(t, struct controller, dfs_cleanup);

	/* Run background CAC here */
	dbg("dfs bgcac timeout - run cleanup\n");
	cntlr_dfs_cleanup(c);

	if (c->cfg.dfs_cleanup_timeout)
		uloop_timeout_set(&c->dfs_cleanup, c->cfg.dfs_cleanup_timeout * 1000);
}

static void cntlr_start(struct uloop_timeout *t)
{
	struct controller *c = container_of(t, struct controller, start_timer);

	if (c->state == CNTLR_INIT) {
		c->state = CNTLR_START;
		cntlr_publish_object(c, "map.controller");
	}
}

static void cntlr_discovery(struct uloop_timeout *t)
{
	struct controller *c = container_of(t, struct controller, discovery_timer);
	struct cmdu_buff *cmdu;

	cmdu = cntlr_gen_ap_autoconfig_search(c, 0x02, 0x00);
	if (!cmdu)
		return;

	c->mid_2g = send_cmdu(c, cmdu);
	cmdu_free(cmdu);

	cmdu = cntlr_gen_ap_autoconfig_search(c, 0x02, 0x01);
	if (!cmdu)
		return;

	c->mid_5g = send_cmdu(c, cmdu);
	cmdu_free(cmdu);

	uloop_timeout_set(t, 180 * 1000);
}

int cntlr_map_sub_cb(void *bus, void *priv, void *data)
{
	struct blob_attr *msg = (struct blob_attr *)data;
	char *str;


	str = blobmsg_format_json(msg, true);
	trace("Received notification '%s'\n", str);
	free(str);

	cntlr_ieee1905_cmdu_event_handler(priv, msg);

	return 0;
}

int cntlr_map_del_cb(void *bus, void *priv, void *data)
{
	struct controller *c = (struct controller *)priv;
	uint32_t *obj = (uint32_t *)data;

	c->subscribed = false;
	fprintf(stdout, "Object 0x%x no longer present\n", *obj);

	return 0;
}

static int controller_subscribe_for_cmdus(struct controller *c)
{
	mapmodule_cmdu_mask_t cmdu_mask = {0};
	uint32_t map_id;
	int ret;


	map_prepare_cmdu_mask(cmdu_mask,
			CMDU_TYPE_TOPOLOGY_DISCOVERY,
			CMDU_TYPE_TOPOLOGY_NOTIFICATION,
			CMDU_TYPE_TOPOLOGY_QUERY,
			CMDU_TYPE_TOPOLOGY_RESPONSE,
			CMDU_TYPE_VENDOR_SPECIFIC,
			CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE,
			CMDU_TYPE_AP_AUTOCONFIGURATION_WSC,
			CMDU_1905_ACK,
			CMDU_BEACON_METRICS_RESPONSE,
			CMDU_AP_METRICS_RESPONSE,
			CMDU_ASSOC_STA_LINK_METRICS_RESPONSE,
			CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE,
			CMDU_CHANNEL_SCAN_REQUEST,
			CMDU_CHANNEL_SCAN_REPORT,
			CMDU_CLIENT_DISASSOCIATION_STATS,
			CMDU_ASSOCIATION_STATUS_NOTIFICATION,
			CMDU_BACKHAUL_STA_CAPABILITY_QUERY,
			CMDU_BACKHAUL_STA_CAPABILITY_REPORT,
			CMDU_CHANNEL_PREFERENCE_REPORT,
			CMDU_CLIENT_STEERING_BTM_REPORT,
			CMDU_STEERING_COMPLETED,
			CMDU_CHANNEL_SELECTION_RESPONSE,
			CMDU_OPERATING_CHANNEL_REPORT,
			CMDU_AP_CAPABILITY_QUERY,
			CMDU_AP_CAPABILITY_REPORT,
			CMDU_CLIENT_CAPABILITY_REPORT,
			CMDU_HIGHER_LAYER_DATA,
			-1);
	memcpy(c->cmdu_mask, cmdu_mask, sizeof(c->cmdu_mask));

	trace("<----------------------------------- %s\n", __func__);
	for (;;) {
		ret = ubus_lookup_id(c->ubus_ctx, map_plugin, &map_id);
		if (!ret)
			break;

		trace("%s not up yet, sleeping for 2s!\n", map_plugin);
		sleep(2);
	}

	c->map_oid = map_id;

	/* register as client to the map module */
	ret = map_subscribe(c->ubus_ctx,
			    &c->map_oid,
			    "mapcontroller", &cmdu_mask, c,
			    cntlr_map_sub_cb,
			    cntlr_map_del_cb,
			    &c->subscriber);
	if (!ret) {
		c->subscribed = true;
	} else {
		trace("Failed to 'register' with %s (err = %s)\n",
		      map_plugin, ubus_strerror(ret));
	}

	return ret;
}

static void ieee1905_info_cb(struct ubus_request *req, int type,
			     struct blob_attr *msg)
{
	struct controller *c = (struct controller *)req->priv;
	struct blob_attr *tb[1];
	static const struct blobmsg_policy ieee_attrs[1] = {
		[0] = { .name = "ieee1905id", .type = BLOBMSG_TYPE_STRING },
	};

	blobmsg_parse(ieee_attrs, 1, tb, blob_data(msg), blob_len(msg));

	if (tb[0]) {
		char *mac;

		mac = blobmsg_get_string(tb[0]);
		hwaddr_aton(mac, c->almac);
		dbg("almac = " MACFMT "\n", MAC2STR(c->almac));
	}
}

static int cntlr_wait_for_ieee1905(struct controller *c)
{
	uint32_t ieee1905_obj;

	for (;;) {
		int ret;

		ret = ubus_lookup_id(c->ubus_ctx, "ieee1905", &ieee1905_obj);
		if (!ret)
			break;

		trace("%s not up yet, sleeping for 2s!\n", map_plugin);
		sleep(2);
	}

	return ubus_call_object(c, ieee1905_obj, "info", ieee1905_info_cb, c);
}

static int cntlr_ackq_timeout_cb(struct cmdu_ackq *q, struct cmdu_ackq_entry *e)
{
	struct controller *a = container_of(q, struct controller, cmdu_ack_q);
	struct cmdu_buff *cmdu = (struct cmdu_buff *) e->cookie;
	int ret;

	trace("%s: ---> cmdu = %04x to "MACFMT" \n", __func__,
		cmdu_get_type(cmdu), MAC2STR(cmdu->origin));

	if (e->resend_cnt-- > 0) {
		ret = send_cmdu_ubus(a, cmdu);
		if (ret < 0)
			err("%s fail to send cmdu\n", __func__);

		return CMDU_ACKQ_TMO_REARM;
	}

	return CMDU_ACKQ_TMO_DELETE;
}

static void cntlr_ackq_delete_cb(struct cmdu_ackq *q, struct cmdu_ackq_entry *e)
{
	struct cmdu_buff *cmdu = (struct cmdu_buff *) e->cookie;

	trace("%s: ---> cmdu = %04x to "MACFMT" \n", __func__,
		cmdu_get_type(cmdu), MAC2STR(cmdu->origin));

	cmdu_free(cmdu);
}

static void uobj_add_event_handler(void *cntlr, struct blob_attr *msg)
{
	char path[32] = {0};
	uint32_t id = 0;
	struct controller *c = (struct controller *) cntlr;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "id", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "path", .type = BLOBMSG_TYPE_STRING }
	};

	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(path, blobmsg_data(tb[1]), sizeof(path) - 1);
	id = (uint32_t) blobmsg_get_u32(tb[0]);
	dbg("|%s:%d| path = [%s] id = [%d] [%u]\n", __func__, __LINE__, path,
			id, id);
	if (!strncmp(path, map_plugin, strlen(map_plugin))) {
		/* TODO: how to handle failure? */
		controller_subscribe_for_cmdus(c);
	}
}


static void cntlr_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev,
		const char *type, struct blob_attr *msg)
{
	int i;
	char *str;
	struct controller *c = container_of(ev, struct controller, evh);
	struct wifi_ev_handler {
		const char *ev_type;
		void (*handler)(void *ctx, struct blob_attr *ev_data);
	} evs[] = {
		{ "ubus.object.add", uobj_add_event_handler }
	};

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	info("[ &controller = %p ] Received [event = %s]  [val = %s]\n",
			c, type, str);

	for (i = 0; i < ARRAY_SIZE(evs); i++) {
		if (!strcmp(type, evs[i].ev_type))
			evs[i].handler(c, msg);
	}

	free(str);
}


void run_controller(void)
{
	struct controller *c;
	struct ubus_context *ctx;
	/* struct ubus_event_handler *ev; */
	sigset_t base_mask;

	sigemptyset(&base_mask);
	sigaddset(&base_mask, SIGHUP);

	sigprocmask(SIG_SETMASK, &base_mask, NULL);
	set_sighandler(SIGPIPE, SIG_IGN);

	c = calloc(1, sizeof(struct controller));
	if (!c)
		return;

	cntlr_dbg("Starting wifi_cntlr... (&cntlr = %p)\n", c);

	uloop_init();
	ctx = ubus_connect(ubus_socket);
	if (!ctx) {
		err("Failed to connect to ubus\n");
		free(c);
		return;
	}
	c->ubus_ctx = ctx;
	INIT_LIST_HEAD(&c->stalist);
	INIT_LIST_HEAD(&c->nodelist);
	INIT_LIST_HEAD(&c->bcnreqlist);
	c->num_nodes = 0;
	INIT_LIST_HEAD(&c->linklist);
	as_init_table(&c->as_table);

	cmdu_ackq_init(&c->cmdu_ack_q);
	c->cmdu_ack_q.timeout_cb = cntlr_ackq_timeout_cb;
	c->cmdu_ack_q.delete_cb = cntlr_ackq_delete_cb;

	ubus_add_uloop(ctx);

	cntlr_wait_for_ieee1905(c);

	cntlr_config_defaults(c, &c->cfg);

	cntlr_resync_config(c, false);

	if (!c->cfg.enabled)
		goto out_exit;

	{
		/* TODO: diff always 1 after first round, will cause failures on
		 * first reload */
		struct node_policy *np;

		list_for_each_entry(np, &c->cfg.nodelist, list) {
			struct node *n;

			np->is_policy_diff = false;
			n = alloc_node_init(c, np->agent_id);
			if (!n)
				goto out_exit;

			n->np = np;

		}
	}


	//cntlr_register_events(c);

	c->state = CNTLR_INIT;
	c->discovery_timer.cb = cntlr_discovery;
	c->start_timer.cb = cntlr_start;
	c->heartbeat.cb = cntlr_periodic_run;
	c->radar_timer.cb = cntlr_radar_exit;
	c->signal_handler.cb = cntlr_signal_periodic_run;
	c->query_nodes.cb = cntlr_query_nodes;
	c->acs.cb = cntlr_acs_run;
	c->dfs_cleanup.cb = cntlr_dfs_cleanup_run;

	uloop_timeout_set(&c->heartbeat, 5 * 1000);
	uloop_timeout_set(&c->discovery_timer, 0);
	uloop_timeout_set(&c->start_timer, waitext ? 5 * 1000 : 0);
	uloop_timeout_set(&c->signal_handler, 5 * 1000);
	uloop_timeout_set(&c->query_nodes, 60 * 1000);

	if (c->cfg.acs_timeout)
		uloop_timeout_set(&c->acs, c->cfg.acs_timeout * 1000);
	if (c->cfg.dfs_cleanup_timeout)
		uloop_timeout_set(&c->dfs_cleanup, c->cfg.dfs_cleanup_timeout * 1000);

	c->evh.cb = cntlr_event_handler;
	ubus_register_event_handler(ctx, &c->evh, "ubus.object.*");

	controller_subscribe_for_cmdus(c);

	/* steer-control */
	INIT_LIST_HEAD(&c->sclist);
	cntlr_load_steer_modules(c);
	if (!list_empty(&c->sclist))
		cntlr_assign_steer_module_default(c);

	uloop_run();

out_exit:
	map_unsubscribe(ctx, c->subscriber);
	cntlr_clean_nodelist(c);
	ubus_unregister_event_handler(ctx, &c->evh);
	cntlr_remove_object(c);
	cmdu_ackq_free(&c->cmdu_ack_q);
	cntlr_config_clean(&c->cfg);
	uloop_done();
	free(c);
}
